#include <iostream>
#include <string>
#include <utility>
#include <vector>
#include <climits> // For INT_MAX
#include <cmath>   // For std::abs, std::floor
#include <algorithm> // For std::min and std::max
#include <HPModel.h>
#include <Utilities.h>
#include <Application.h>

struct Block {
    int startIndex;                 // Start index in the sequence
    int endIndex;                   // End index in the sequence
    std::string type;               // "x", "y", or "z"
    std::string residues;           // Subsequence of residues
    std::string foldConnector;      // The very first fold of the block, this depends on the previous blocks, so can not be set immidiately
    std::string normalFormFolds;    // The relative folds needed for this block to be in normal form, this excludes the very first fold!
};

std::string connectorFoldsEven(const std::string &residues) {
    int size = residues.length();
    if (size == 0) {
        return "";
    }
    if (size == 2) {
        return "f";
    }
    if (size == 4) {
        return "frr";
    }
    int forward = (size / 2) - 1;
    std::string folds;
    for (int i = 0; i < forward; i++) {
        folds += "f";
    }
    folds += "rr";
    for (int i = 0; i < forward; ++i) {
        folds += "f";
    }
    return folds;
}

std::string connectorFoldsOdd(const std::string &residues) {
    int size = residues.length();
    if (size == 1) {
        return "";
    }
    if (size == 3) {
        return "rr";
    }
    int forward = (size - 3) / 2;
    std::string folds;
    for (int i = 0; i < forward; ++i) {
        folds += "f";
    }
    folds += "rr";
    for (int i = 0; i < forward; ++i) {
        folds += "f";
    }
    return folds;
}

std::string getConnectorFold(const std::string &residues) {
    // Note again that this does not include the first fold as this is determined by the connection
    int size = residues.length();
    // There are situations (usually with 1 length blocks) where it is possible to push the current off type block into
    // the following bloc, thus resulting in an odd length connector! This has to be handled seperately
    if (size % 2 == 1) {
        return connectorFoldsOdd(residues);
    } else {
        return connectorFoldsEven(residues);
    }
}

std::string getConnectorFold3d(const std::string &residues) {
    int size = residues.length();
    if (size == 2) {
        // TODO: This requires some finagling when we set the connector
        return "u";
    }
    if (size == 4) {
        return "fuu";
    }
    int forward = (size / 2) - 1;
    std::string folds;
    for (int i = 0; i < forward - 1; i++) {
        folds += "f";
    }
    folds += "uu";
    for (int i = 0; i < forward; i++) {
        folds += "f";
    }
    return folds;
}


std::string foldMiniBlock(const std::string &residues) {
    // Assume "mini block" is a sequence of 1(0^x), for x odd
    // We should be able to assume that we are always facing "forward" compared to the rest of the blocks
    std::string folds;
    auto size = residues.length();
    if (size <= 1) {
        return "";
    }
    if (size == 2) {
        return "f";
    }
    if (size == 4) {
        return "lrr";
    }
    int forward = (size / 2) - 1;
    folds += "l";
    for (int i = 0; i < forward - 1; i++) {
        folds += "f";
    }
    folds += "rr";
    for (int i = 0; i < forward - 1; i++) {
        folds += "f";
    }
    return folds;
}

std::vector<std::string> splitIntoMiniBlocks(const std::string &residues) {
    std::vector<std::string> result{};
    std::string workingString;
    bool empty = true;
    for (auto c: residues) {
        if (c == 'h') {
            if (empty) {
                workingString += c;
                empty = false;
                continue;
            }
            // String is not empty
            result.push_back(workingString);
            workingString = c;
            continue;
        }
        // c != 'h' <=> c == 'p'
        workingString += c;
    }
    result.push_back(workingString);
    return result;
}

std::string limitHinSequenceByJ(const std::string& residues, int J) {
    std::string result = "";

    int numberOfHFound = 0;
    for (int i = 0; i < residues.length(); i++) {
        char c = residues[i];
        if (c == 'p') {
            result += c;
            continue;
        }
        if (i == 0) {
            result += c;
            continue;
        }
        if (i == residues.length() - 1) {
            result += c;
            continue;
        }

        // Check if we have room for more H's
        if (numberOfHFound < J - 2) {
            result += c;
            numberOfHFound++;
        } else {
            result += 'p';
        }
    }
    return result;
}

int calculatePotentialPairsInString(std::string residues) {
    int pairsFound = 0;
    int gapSize = 0;
    for (int i = 0; i < residues.length(); i++) {
        char c = residues.at(i);
        if (c == 'p') {
            gapSize++;
            continue;
        }
        if (pairsFound == 0) {
            pairsFound++;
            gapSize = 0;
            continue;
        }
        if (gapSize % 2 == 1) {
            pairsFound++;
            gapSize = 0;
            continue;
        }
        gapSize++;
    }
    return pairsFound;
}

// Always assume normal folds fold LEFT (relative) - based on HeartIStrail1995
std::string calculateNormalFormFolds(const Block &block, bool simpleCase, int J) {
    // TODO: Handle special case of block of size 1
    int blockLength = block.endIndex - block.startIndex;
    bool isConnector = true;
    for (auto c: block.residues) {
        if (c == 'h') {
            isConnector = false;
        }
    }
    std::string residues = limitHinSequenceByJ(block.residues, J);
    if (isConnector) {
        if (simpleCase) {
            return getConnectorFold(block.residues);
        } else {
            return getConnectorFold3d(block.residues);
        }
    }

    std::vector<std::string> miniBlocks = splitIntoMiniBlocks(residues);
    std::string folds;
    for (int i = 0; i < miniBlocks.size(); i++) {
        if (i == 0) {
            // Special case for first mini block is that we do not know what the connection string is, so we have to subtract that
            std::string fullMiniFoldString = foldMiniBlock(miniBlocks.at(i));
            //std::string reducedMiniFoldString = fullMiniFoldString.substr(1, fullMiniFoldString.length() - 1);
            //folds += reducedMiniFoldString;
            folds += fullMiniFoldString;
            continue;
        }
        if (i == miniBlocks.size() - 1) {
            // The very last block will always be exactly "h"
            std::string previousFold = foldMiniBlock(miniBlocks.at(i - 1));
            // Special case where the previous fold is small enough that there is no extrusion
            std::string previousFoldLastChar = previousFold.substr(previousFold.length() - 1, 1);
            if (previousFold.size() <= 2) {
                folds += "f";
            } else {
                if (previousFoldLastChar == "l") {
                    folds += "f";
                } else {
                    folds += "l";
                }
            }
            continue; // Effectively a break
        }
        // Not the final block and not the first block :)
        // TODO: This could be cached, but these strings are so small it probably doesn't matter
        std::string previousFoldString = foldMiniBlock(miniBlocks.at(i - 1));
        std::string miniFoldString = foldMiniBlock(miniBlocks.at(i));
        // At this point we know the connection string for all the mini blocks!
        if (previousFoldString.size() <= 2) {
            folds += "f";
        } else {
            folds += "l";
        }
        folds += miniFoldString;
    }
    return folds;
}

std::string getFoldsForBlockNormalForm(const Block &block) {
    if (block.foldConnector == "NONE") {
        return block.normalFormFolds;
    }
    if (block.foldConnector.empty()) {
        //return block.normalFormFolds;
        throw std::invalid_argument("Cannot get folds for a block before the fold connector is set correctly!\n");
    }
    return block.foldConnector + block.normalFormFolds;
}

// Helper function to get the length of a 'p' sequence starting at index i
int getPSequenceLength(const std::string &sequence, int start) {
    int count = 0;
    int n = sequence.length();
    while (start + count < n && sequence[start + count] == 'p') {
        count++;
    }
    return count;
}

std::vector<Block> decomposeSequence(const std::string &sequence) {
    std::vector<Block> blocks;
    int n = sequence.length();
    int i = 0;

    // Handle possible initial even-length 'p' sequence (z0)
    if (sequence[i] == 'p') {
        int p_len = getPSequenceLength(sequence, i);
        Block separator;
        separator.startIndex = i;
        separator.endIndex = i + p_len - 1;
        separator.residues = sequence.substr(i, p_len);
        blocks.push_back(separator);
        i += p_len;
    }
    bool nextBlockMustBeSeperator = false;
    while (i < n) {
        if (sequence[i] == 'h') {
            // Start of a new hydrophobic block
            int blockStart = i;
            int blockEnd = i;

            while (true) {
                i++; // Move past the 'h'
                // Check for 'p's
                if (i >= n || sequence[i] != 'p') {
                    break;
                }
                int pStart = i;
                int pLen = getPSequenceLength(sequence, i);
                // Check if 'p' sequence is odd-length
                if (pLen % 2 == 1) {
                    // Include the 'p's in the block
                    i += pLen;
                    // Check for another 'h'
                    if (i < n && sequence[i] == 'h') {
                        blockEnd = i; // Include the 'h' in the block
                        continue;      // Continue the loop to check for more
                    } else {
                        break; // No more 'h's, end of block
                    }
                } else {
                    // Even-length 'p's are separators
                    break;
                }
            }

            // Create hydrophobic block
            /*
            if (nextBlockMustBeSeperator) {
                // To avoid xy and yx combinations, we have to insert a 0 length z block
                Block zeroLength;
                zeroLength.startIndex = -1;
                zeroLength.endIndex = -1;
                zeroLength.residues = "";
                blocks.push_back(zeroLength);
            }
             */
            Block block;
            block.startIndex = blockStart;
            block.endIndex = blockEnd;
            block.residues = sequence.substr(blockStart, blockEnd - blockStart + 1);

            blocks.push_back(block);
            nextBlockMustBeSeperator = true;
        }

        // Handle separators (even-length 'p's)
        if (i < n && sequence[i] == 'p') {
            int pStart = i;
            int pLen = getPSequenceLength(sequence, i);

            // Even-length 'p's are separators
            Block separator;
            separator.startIndex = pStart;
            separator.endIndex = pStart + pLen - 1;
            separator.residues = sequence.substr(pStart, pLen);
            blocks.push_back(separator);
            nextBlockMustBeSeperator = false;

            i += pLen;
        }
    }

    return blocks;
}

void assignBlockTypes(std::vector<Block> &blocks) {
    int hydrophobicBlockCount = 0;
    for (size_t j = 0; j < blocks.size(); ++j) {
        if (blocks[j].residues[0] == 'h') {
            hydrophobicBlockCount++;
            blocks[j].type = (hydrophobicBlockCount % 2 == 1) ? "y" : "x";
        } else {
            blocks[j].type = "z"; // Separator (even-length polar residues)
        }
    }
}

void calculateCounts(const std::vector<Block> &blocks, int &N_x, int &N_y) {
    N_x = 0;
    N_y = 0;
    for (const auto &block: blocks) {
        if (block.type == "x") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_x++;
                }
            }
        } else if (block.type == "y") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_y++;
                }
            }
        }
    }
}

void partitionSequence(const std::vector<Block> &blocks,
                       std::vector<Block> &B_prime,
                       std::vector<Block> &B_double_prime) {
    size_t k = blocks.size();

    // Initialize e and E
    int e = -1, E = INT_MAX; // Best partition metrics

    // Define cond(a, b, A, B) ≡ (a < b) || ((a == b) && (A < B))
    auto cond = [](int a, int b, int A, int B) {
        return (a < b) || ((a == b) && (A < B));
    };

    // **Initial Partition**: B1 = z0 b1 z1, B2 = b2 z2 ... bk zk
    std::vector<Block> B1(blocks.begin(), blocks.begin() + 3); // First 3 blocks
    std::vector<Block> B2(blocks.begin() + 3, blocks.end());   // Remaining blocks

    // Calculate N_x(B1), N_y(B1), N_x(B2), N_y(B2)
    int N_x_B1 = 0, N_y_B1 = 0, N_x_B2 = 0, N_y_B2 = 0;
    calculateCounts(B1, N_x_B1, N_y_B1);
    calculateCounts(B2, N_x_B2, N_y_B2);

    // Compute m_{xy}, M_{xy}, m_{yx}, M_{yx}
    int m_xy = std::min(N_x_B1, N_y_B2);
    int m_yx = std::min(N_y_B1, N_x_B2);
    int M_xy = std::max(N_x_B1, N_y_B2);
    int M_yx = std::max(N_y_B1, N_x_B2);

    // **Initial Assignment** based on m_{xy} vs m_{yx}
    if (m_xy > m_yx) {
        B_prime = B2;
        B_double_prime = B1;
        e = m_xy;
        E = M_xy;
    } else {
        B_prime = B1;
        B_double_prime = B2;
        e = m_yx;
        E = M_yx;
    }

    // **Loop** over possible partition points starting from index 3
    for (size_t i = 3; i <= k; ++i) {
        // Update B1 and B2 for current partition
        B1 = std::vector<Block>(blocks.begin(), blocks.begin() + i);
        B2 = std::vector<Block>(blocks.begin() + i, blocks.end());

        // Recalculate counts for current B1 and B2
        calculateCounts(B1, N_x_B1, N_y_B1);
        calculateCounts(B2, N_x_B2, N_y_B2);

        // Recompute m_{xy}, M_{xy}, m_{yx}, M_{yx}
        m_xy = std::min(N_x_B1, N_y_B2);
        m_yx = std::min(N_y_B1, N_x_B2);
        M_xy = std::max(N_x_B1, N_y_B2);
        M_yx = std::max(N_y_B1, N_x_B2);

        // **Check cond(e, m_{xy}, E, M_{xy})**
        if (cond(e, m_xy, E, M_xy)) {
            B_prime = B2;
            B_double_prime = B1;
            e = m_xy;
            E = M_xy;
        }

        // **Check cond(e, m_yx, E, M_yx)**
        if (cond(e, m_yx, E, M_yx)) {
            B_prime = B1;
            B_double_prime = B2;
            e = m_yx;
            E = M_yx;
        }
    }
}

// Directions for 3D movement using vectors
struct Vector3D {
    int x, y, z;

    Vector3D(int x_, int y_, int z_) : x(x_), y(y_), z(z_) {}

    // Rotate around the specified axis by 90 degrees
    void rotate(char axis, int angle) {
        angle = angle % 360;
        if (angle < 0) {
            angle += 360;
        }

        int temp;
        switch (axis) {
            case 'X':
            case 'x':
                for (int i = 0; i < angle; i += 90) {
                    temp = y;
                    y = -z;
                    z = temp;
                }
                break;
            case 'Y':
            case 'y':
                for (int i = 0; i < angle; i += 90) {
                    temp = x;
                    x = z;
                    z = -temp;
                }
                break;
            case 'Z':
            case 'z':
                for (int i = 0; i < angle; i += 90) {
                    temp = x;
                    x = -y;
                    y = temp;
                }
                break;
            default:
                throw std::invalid_argument("Invalid axis for rotation");
        }
    }
};

struct Point3D {
    int x, y, z;
};

void moveForward(Point3D &pos, Vector3D dir) {
    pos.x += dir.x;
    pos.y += dir.y;
    pos.z += dir.z;
}

// Map fold characters to 3D directions
Vector3D getDirection3D(char fold, Vector3D currentDir) {
    Vector3D newDir = currentDir;
    switch (fold) {
        case 'f':
            // Move forward in current direction
            break;
        case 'l':
            // Rotate 90 degrees to the left around the Y-axis
            newDir.rotate('Y', -90);
            break;
        case 'r':
            // Rotate 90 degrees to the right around the Y-axis
            newDir.rotate('Y', 90);
            break;
        case 'u':
            // Rotate 90 degrees upward around the X-axis
            newDir.rotate('X', -90);
            break;
        case 'd':
            // Rotate 90 degrees downward around the X-axis
            newDir.rotate('X', 90);
            break;
        default:
            throw std::invalid_argument("Invalid fold character");
    }
    return newDir;
}

std::string generateFoldsForBlocks(const std::vector<Block> &blocks,
                                   const std::string &superblockType,
                                   std::vector<Point3D> &positions,
                                   Vector3D &currentDir,
                                   bool isFirstBlock,
                                   int &residuesPlaced,
                                   int totalResidues) {
    std::string folds;
    Point3D currentPos = positions.back();

    bool isFirstResidue = isFirstBlock && positions.size() == 1;

    for (const auto &block: blocks) {
        size_t num_residues = block.residues.length();

        for (size_t i = 0; i < num_residues; ++i) {
            char residue = block.residues[i];

            if (isFirstResidue) {
                // Place the first residue without moving
                isFirstResidue = false;
                residuesPlaced++;
                continue; // Skip generating a fold for the first residue
            }

            if (residuesPlaced >= totalResidues) {
                // All residues have been placed; do not add more folds
                break;
            }

            // Determine the fold based on block type and position
            char fold;
            if (block.type == superblockType) {
                // Move forward along the face
                fold = 'f';
            } else if (block.type == "z") {
                // Alternate up and down moves for 'z' blocks
                fold = ((residuesPlaced % 2) == 0) ? 'u' : 'd';
            } else {
                // Alternate left and right turns for other blocks
                fold = ((residuesPlaced % 2) == 0) ? 'l' : 'r';
            }

            // Append the fold instruction
            folds += fold;

            // Update the current direction
            currentDir = getDirection3D(fold, currentDir);

            // Update position
            moveForward(currentPos, currentDir);
            positions.push_back(currentPos);

            // Increment residues placed
            residuesPlaced++;
        }
    }

    return folds;
}

std::string calculateFoldConnector2d(Block &previousBlock, Block &currentBlock, std::string superBlockType) {
    int previousSize = previousBlock.residues.length() - 1;
    std::string previousFoldLastChar = "NONE";
    if (previousSize > 0) {
        previousFoldLastChar = previousBlock.normalFormFolds.substr(
                previousBlock.normalFormFolds.length() - 1, 1);
    }
    // TODO: Update block type correctly when combining blocks
    // The hacky way we check if a block is a combination block is to see if the block type is empty, which should work?
    if (currentBlock.type.empty()) {
        // If we are in a combination block, there are two cases (similar to other blocks) where we have either 3+ (uneven)
        // residues, or we have exactly *1* residue
        if (currentBlock.residues.length() == 1) {
            return "f";
        }
        return "l";
    }
    // Same behaviour if the previous block is a combination block (we need 2 left folds to counter the 2 right folds)
    if (previousBlock.type.empty()) {
        if (previousBlock.residues.length() == 1) {
            return "f";
        }
        return "l";
    }
    if (previousSize <= 2) {
        return "f";
    }
    // I believe this is because this will (hopefully) always
    if (previousFoldLastChar == "l") {
        return "f";
    }
    return "l";
}

// Return the updated vectors
std::tuple<std::vector<Block>, std::vector<Block>>
updateConnectionFoldForSuperBlocksSimpleCase(std::vector<Block> B_prime, std::vector<Block> B_doublePrime) {
    std::vector<Block> newBPrime{};
    std::vector<Block> newBDoublePrime{};
    // For the very first block of B_prime we can assume that we are facing forward, thus the first connector is left
    Block firstBlock = B_prime.at(0);
    //firstBlock.foldConnector = "l";
    firstBlock.foldConnector = "NONE";
    newBPrime.push_back(firstBlock);
    for (int i = 1; i < B_prime.size(); i++) {
        // We can now assume that a previous block always exists
        Block previousBlock = B_prime.at(i - 1);
        Block currentBlock = B_prime.at(i);
        int previousSize = previousBlock.residues.length() - 1;
        std::string previousFoldLastChar = "NONE";
        currentBlock.foldConnector = calculateFoldConnector2d(previousBlock, currentBlock, "y");
        /*
        if (previousSize > 0) {
            previousFoldLastChar = previousBlock.normalFormFolds.substr(
                    previousBlock.normalFormFolds.length() - 1, 1);
        }
        if (previousSize <= 2) {
            currentBlock.foldConnector = "f";
        } else {
            if (previousFoldLastChar == "l") {
                currentBlock.foldConnector = "f";
            } else {
                currentBlock.foldConnector = "l";
            }
        }
         */
        // TODO: This is likely wrong, but bruteforce solution
        newBPrime.push_back(currentBlock);
    }

    Block doublePrimeFirstBlock = B_doublePrime.at(0);
    // TODO: This is likely wrong, consider special cases for connecting the two super blocks
    //doublePrimeFirstBlock.foldConnector = "NONE";
    // This is "easy" if a connecting block of at least size 2, as this is entirely handled within that block - probably :D
    if (doublePrimeFirstBlock.type == "z" && doublePrimeFirstBlock.residues.size() >= 2) {
        doublePrimeFirstBlock.foldConnector = "f";
    } else {
        doublePrimeFirstBlock.foldConnector = "r";
    }
    newBDoublePrime.push_back(doublePrimeFirstBlock);
    for (int i = 1; i < B_doublePrime.size(); i++) {
        Block previousBlock = B_doublePrime.at(i - 1);
        Block currentBlock = B_doublePrime.at(i);
        int previousSize = previousBlock.residues.length() - 1;
        if (i == 1) {
            if (!currentBlock.type.empty()) {
                currentBlock.foldConnector = "r";
            } else {
                currentBlock.foldConnector = "f";
            }
        } else {
            currentBlock.foldConnector = calculateFoldConnector2d(previousBlock, currentBlock, "x");
        }
        /*
        else if (previousSize <= 2) {
            currentBlock.foldConnector = "f";
        } else {
            currentBlock.foldConnector = "l";
        }
         */

        newBDoublePrime.push_back(currentBlock);
    }

    return std::make_tuple(newBPrime, newBDoublePrime);
}

std::string getCombinedFolds(std::vector<Block> B_prime, std::vector<Block> B_double_prime) {
    // TODO: Consider the final folding thingy
    std::string folds;
    for (auto block: B_prime) {
        std::string foldsToAdd = getFoldsForBlockNormalForm(block);
        folds += foldsToAdd;
        std::cout << "Found " << foldsToAdd << "(" << foldsToAdd.size() << ") folds to add for block " << block.residues
                  << " (" << block.residues.length() << ")\n";
    }
    //folds += "f";
    for (auto block: B_double_prime) {
        std::string foldsToAdd = getFoldsForBlockNormalForm(block);
        folds += foldsToAdd;
        std::cout << "Found " << foldsToAdd << "(" << foldsToAdd.size() << ") folds to add for block " << block.residues
                  << " (" << block.residues.length() << ")\n";
    }
    return folds;
}

std::string foldConnectingSeperator(std::string residues) {
    // TODO: For now we only fold as a simple 'U' shape, not considering edge cases of scoring points at the mid-point fold - consider HeartIStrail1995 figure 6 approach
    // Note that this "connector" MIGHT be of odd length
    int midPoint = residues.length() / 2 - 2;
    if (residues.size() <= 1) {
        return "";
    }
    if (residues.size() == 2) {
        return "r";
    }
    std::string resultingFolds;
    for (int i = 0; i < midPoint; ++i) {
        resultingFolds += "f";
    }
    resultingFolds += "rr";
    for (int i = 0; i < midPoint - 1; ++i) {
        resultingFolds += "f";
    }
    return resultingFolds;
}

Block appendBlock(const Block &originalBlock, const Block &toAppend) {
    Block resBlock = originalBlock;
    resBlock.residues += toAppend.residues;
    resBlock.normalFormFolds += toAppend.normalFormFolds;
    if (resBlock.residues.size() % 2 == 0) {
        resBlock.type = "z";
    } else {
        resBlock.type = "";
    }
    return resBlock;
}

std::vector<Block> calculateNormalFoldsForSuperBlock(std::vector<Block> superBlock, std::string superBlockType) {
    std::vector<Block> result{};
    Block offTypeBlock = Block{};
    bool offTypeBlockExists = false; // Needed because of zero-length separators
    for (int i = 0; i < superBlock.size(); i++) {
        Block currentBlock = superBlock.at(i);
        std::string currentBlockType = currentBlock.type;
        // Special case if the first block is a seperator, it means that we are in the beginning of the second superblock and need to fold accordingly
        if (i == 0 && currentBlock.type == "z") {
            // It might be the case that we have built up a larger seperator block at this point, so that whole block should be folded if it exists
            offTypeBlock = appendBlock(offTypeBlock, currentBlock);
            offTypeBlock.normalFormFolds = foldConnectingSeperator(offTypeBlock.residues);
            result.push_back(offTypeBlock);
            // Shouldn't be needed, but better safe than sorry, the off type block should be cleared after every use
            offTypeBlock = Block{};
            offTypeBlockExists = false;
            continue;
        }
        // Similar special case if the last block is a seperator, only one of these cases should be possible in a pair of superblocks
        if (i == superBlock.size() - 1 && currentBlock.type == "z") {
            offTypeBlock = appendBlock(offTypeBlock, currentBlock);
            offTypeBlock.normalFormFolds = foldConnectingSeperator(offTypeBlock.residues);
            result.push_back(offTypeBlock);
            offTypeBlock = Block{};
            offTypeBlockExists = false;
            continue;
        }
        bool blockTypeMatch = currentBlockType == superBlockType;
        // In general, for an x block in an x-superblock, it can always be folded alone without any further consideration
        // Symmetric for y blocks in y-superblocks
        if (blockTypeMatch) {
            // If our combination block (zyz/yz for x-superblock) is not empty, we need to fold that as a connector first, before we consider the current block
            if (offTypeBlockExists) {
                offTypeBlock.normalFormFolds = getConnectorFold(offTypeBlock.residues);
                result.push_back(offTypeBlock);
                offTypeBlock = Block{};
                offTypeBlockExists = false;
            }

            currentBlock.normalFormFolds = calculateNormalFormFolds(currentBlock, true, 0);
            // TODO: HACKY WRONG SOLUTION, FIX
            currentBlock.foldConnector = "l";
            result.push_back(currentBlock);
            continue;
        }
            // for y and z blocks in x-superblocks we need to concatenate them and fold them as a seperator
        else {
            // We know at this point that this cannot be the final block - would be caught earlier
            // The reason we are allowed to simply do this, is that there are two possible cases:
            // 1) This is a combination block zyz/yz followed by an x block, in which case this i handled in the previous case
            // 2) This is a combination block zyz/yz not followed by an x block, in which case this is handled in the last block case
            offTypeBlock = appendBlock(offTypeBlock, currentBlock);
            offTypeBlockExists = true;
            continue;
        }

        throw std::invalid_argument("Unknown case in calculateNormalFoldsForSuperBlock!\n");
    }
    return result;
}

std::string applyAlgorithmB(std::vector<Block> &B_prime,
                            std::vector<Block> &B_double_prime,
                            const std::string &sequence) {
    // Initialize variables for folding logic
    std::string folds;
    Vector3D currentDir(1, 0, 0); // Start moving along positive X-axis
    Point3D currentPos = {0, 0, 0};
    std::vector<Point3D> positions;
    positions.push_back(currentPos);

    // Total number of residues
    int totalResidues = sequence.length();

    // Counter for placed residues (start from 0 since the first residue is at the origin)
    int residuesPlaced = 0;

    // Calculate N_x(B'') and N_y(B')
    int N_x_B_double_prime = 0, N_y_B_prime = 0;

    // Calculate N_x(B'')
    for (const auto &block: B_double_prime) {
        if (block.type == "x") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_x_B_double_prime++;
                }
            }
        }
    }

    // Calculate N_y(B')
    for (const auto &block: B_prime) {
        if (block.type == "y") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_y_B_prime++;
                }
            }
        }
    }

    // Locate separator \( z_i \)
    Block z_i;
    bool z_i_found = false;
    for (const auto &block: B_double_prime) {
        if (block.type == "z") {
            z_i = block;
            z_i_found = true;
            break;
        }
    }

    // Apply the conditional steps based on the face conditions
    // As per HeartIstrail1995 - b' is always a y-superblock and b'' is always an x-superblock
    auto foldedBPrime = calculateNormalFoldsForSuperBlock(B_prime, "y");
    B_prime.assign(foldedBPrime.begin(), foldedBPrime.end());
    auto foldedBDoublePrime = calculateNormalFoldsForSuperBlock(B_double_prime, "x");
    B_double_prime.assign(foldedBDoublePrime.begin(), foldedBDoublePrime.end());

    if (N_y_B_prime == N_x_B_double_prime || (z_i_found && z_i.residues.length() > 0)) {
        auto updateTuple = updateConnectionFoldForSuperBlocksSimpleCase(B_prime, B_double_prime);
        auto newBPrime = std::get<0>(updateTuple);
        auto newBDoublePrime = std::get<1>(updateTuple);
        std::string folds = getCombinedFolds(newBPrime, newBDoublePrime);
        return folds;

        // Standard folding point with \( z_i \)
    } else if (N_y_B_prime > N_x_B_double_prime && (!z_i_found || z_i.residues.length() == 0)) {
        // Redefine \( B' \) to exclude the first residue
        auto redefined_B_prime = B_prime;
        if (!redefined_B_prime.empty() && !redefined_B_prime.front().residues.empty()) {
            redefined_B_prime.front().residues.erase(0, 1); // Exclude the first residue
        }
        folds += generateFoldsForBlocks(B_double_prime, "x", positions, currentDir, true, residuesPlaced,
                                        totalResidues);
        folds += generateFoldsForBlocks(redefined_B_prime, "y", positions, currentDir, false, residuesPlaced,
                                        totalResidues);
    } else if (N_y_B_prime < N_x_B_double_prime && (!z_i_found || z_i.residues.length() == 0)) {
        // Redefine \( B'' \) to exclude the last residue
        auto redefined_B_double_prime = B_double_prime;
        if (!redefined_B_double_prime.empty() && !redefined_B_double_prime.back().residues.empty()) {
            redefined_B_double_prime.back().residues.pop_back(); // Exclude the last residue
        }
        folds += generateFoldsForBlocks(redefined_B_double_prime, "x", positions, currentDir, true, residuesPlaced,
                                        totalResidues);
        folds += generateFoldsForBlocks(B_prime, "y", positions, currentDir, false, residuesPlaced, totalResidues);
    }

    // Handle the face conditions, ensuring that folds correspond to amino acid placements
    if (residuesPlaced < totalResidues) {
        // We have residues left to place, so we need to decide how to fold them
        int residuesLeft = totalResidues - residuesPlaced;

        if (N_y_B_prime == N_x_B_double_prime) {
            // Case (a): Refold the last residue
            // Since we have at least one residue left, we can fold it
            folds += 'f'; // Move forward
            currentDir = getDirection3D('f', currentDir);
            moveForward(currentPos, currentDir);
            positions.push_back(currentPos);
            residuesPlaced++;
        } else if (std::abs(N_y_B_prime - N_x_B_double_prime) > 1) {
            // Case (b): Wrap longer superblock around shorter
            // Use remaining residues to create the folds
            char foldsSequence[] = {'u', 'f', 'd'};
            for (char fold: foldsSequence) {
                if (residuesPlaced >= totalResidues) break;
                folds += fold;
                currentDir = getDirection3D(fold, currentDir);
                moveForward(currentPos, currentDir);
                positions.push_back(currentPos);
                residuesPlaced++;
            }
        } else {
            // Case (c): Modify fold of longer face
            char foldsSequence[] = {'r', 'f', 'l'};
            for (char fold: foldsSequence) {
                if (residuesPlaced >= totalResidues) break;
                folds += fold;
                currentDir = getDirection3D(fold, currentDir);
                moveForward(currentPos, currentDir);
                positions.push_back(currentPos);
                residuesPlaced++;
            }
        }
    }

    // Ensure we have exactly n - 1 folds
    if (folds.length() > totalResidues - 1) {
        folds = folds.substr(0, totalResidues - 1);
    }
    return folds;
}


// Only consider type and residues, should always be called BEFORE foldConnector/normalFormFolds are calculated
Block combineBlocks(const Block& b1, const Block& b2) {
    Block result = {};
    result.type = b1.type;
    result.residues = b1.residues + b2.residues;
    return result;
}

// TODO: maybe look for a working implementation here, but I think in general this approach is a red herring
//  - I don't think it's actually possible to combine blocks without losing parity
//  A potential alternative approach is to lengthen the seperator instead of the x/y block, meaning we turn seperator:
//  z0 into seperator z0x0z1y0z2 - it is not obvious where this would be benefitial
std::vector<Block> combineBlocksIfTooSmall(std::vector<Block> superBlock, int J) {
    std::vector<Block> result = {};
    Block unfinishedBlock = {};
    int blocksCombined = 0; // Slightly bad name but this is the *number of times* we have combined blocks
    for (int i = 0; i < superBlock.size(); i++) {
        Block currentBlock = superBlock.at(i);
        if (!unfinishedBlock.residues.empty()) {
            // Unfinished block exists
            currentBlock = combineBlocks(unfinishedBlock, currentBlock);
        }
        int currentPairs = calculatePotentialPairsInString(currentBlock.residues);
        if (currentPairs == 0) {
            // If there are no pairs, it must be a seperator, and separators can ONLY be appended
            result.push_back(currentBlock);
            blocksCombined = 0;
            continue;
        }
        if (currentPairs < J) {
            unfinishedBlock = currentBlock;
            blocksCombined++;
        } else {
            result.push_back(currentBlock);
            blocksCombined = 0;
        }
    }
    // Runtime check to ensure that the new block is actually the same size as old block (by residues)
    int previousSize = 0;
    int newSize = 0;
    for (auto block : superBlock) {
        previousSize += block.residues.length();
    }
    for (auto block : result) {
        newSize += block.residues.length();
    }
    if (previousSize != newSize) {
        throw std::runtime_error("Combination block error, mismatching sizes of: " + std::to_string(previousSize) + " and " + std::to_string(newSize) + "\n");
    }
    return result;
}

std::string calculateFoldConnector(Block previousBlock, Block currentBlock, bool foldUp, char currentFoldDirectionChar) {
    if (!currentBlock.foldConnector.empty()) {
        return currentBlock.foldConnector;
    }
    std::string previousFoldLastChar = previousBlock.normalFormFolds.substr(
            previousBlock.normalFormFolds.length() - 1, 1);
    int previousSize = previousBlock.residues.size();

    // Special case for seperator of size 2
    if (currentBlock.normalFormFolds == "u") {
        return "f";
    }
    // Same for upside/down connector
    if (currentBlock.normalFormFolds == "d") {
        return "f";
    }
    // If the previous block was exactly a seperator of size 2, we need to update current block fold connector accordingly
    if (previousBlock.normalFormFolds == "d" || previousBlock.normalFormFolds == "u") {
        // Slightly ugly, but the basic idea here is that the updated normal form folds set in the previous block
        // will not be carried over here, so we have to do the up/down calculation again, but inverted since it
        // would have been set last iteration, and we cannot set it again now!
        if (currentFoldDirectionChar == 'd') {
            return "u";
        } else {
            return "d";
        }
    }    if (previousSize % 2 == 0) {

        // There is a single special case where if we are at the first x/y block of the second part of the
        // superblock, and the previous seperator had exactly length 2, then we must set the fold connector as 'r'
        // to return to a good state
        if (!foldUp && previousBlock.residues.size() == 2) {
            return "r";
        } else {
            // We fold all separators up no matter what, to do this, we want the fold connector to be forward,
            // otherwise we have to do a lot of recovery with the fold direction later
            return "f";
        }
    } else {
        if (previousFoldLastChar == "l") {
            return "f";
        } else if (previousFoldLastChar == "u" || previousFoldLastChar == "d") {
            return "f";
        } else {
            return "l";
        }
    }

   throw std::runtime_error("Unable to determine fold connector D:\n");
}

std::string foldSuperBlock(std::vector<Block> superBlock, int J, int K, bool foldUp) {
    // Ensure that each block in super block is at least length J (or a seperator)
    // Avoid moving superblock here for debug purposes :)
    //std::vector<Block> superBlockMinJLength = combineBlocksIfTooSmall(superBlock, J);
    std::vector<Block> updatedSuperBlock = {};
    // Swap this between "u" and "d" every time we insert a seperator
    char currentFoldDirectionChar = foldUp ? 'u' : 'd';
    if (foldUp) {
        Block firstBlock = superBlock.at(0);
        firstBlock.foldConnector = "NONE";
        firstBlock.normalFormFolds = calculateNormalFormFolds(firstBlock, false, J);
        updatedSuperBlock.push_back(firstBlock);
    } else {
        // If we are not folding up, we need to handle some funny haha connection point here instead :)
        Block firstBlock = superBlock.at(0);
        firstBlock.foldConnector = "f";
        if (firstBlock.residues.size() == 2) {
            firstBlock.normalFormFolds = "r";
        } else {
            // A bit sneaky, but if we consider this a simple case (2d) it will fold this seperator to the right
            // (only works for separators of size > 2)
            firstBlock.normalFormFolds = calculateNormalFormFolds(firstBlock, true, 0);
        }
        updatedSuperBlock.push_back(firstBlock);
    }
    // Most of this approach follows what we do for Algorithm B
    for (int i = 1; i < superBlock.size(); ++i) {
        Block previousBlock = superBlock.at(i - 1);
        previousBlock.normalFormFolds = calculateNormalFormFolds(previousBlock, false, J);
        Block currentBlock = superBlock.at(i);
        currentBlock.normalFormFolds = calculateNormalFormFolds(currentBlock, false, J);
        int previousSize = previousBlock.residues.length() - 1;
        if (currentBlock.normalFormFolds.contains("u")) {
            std::string newNormalFormFolds = currentBlock.normalFormFolds;
            std::replace(newNormalFormFolds.begin(), newNormalFormFolds.end(), 'u', currentFoldDirectionChar);
            currentBlock.normalFormFolds = newNormalFormFolds;
            if (currentFoldDirectionChar == 'u') {
                currentFoldDirectionChar = 'd';
            } else {
                currentFoldDirectionChar = 'u';
            }
        }
        currentBlock.foldConnector = calculateFoldConnector(previousBlock, currentBlock, foldUp, currentFoldDirectionChar);
        updatedSuperBlock.push_back(currentBlock);
    }

    std::string folds;
    for (auto block: updatedSuperBlock) {
        std::string foldsToAdd = getFoldsForBlockNormalForm(block);
        folds += foldsToAdd;
    }
    return folds;
}

/**
 * Function for splitting up B' and B'' into smaller super blocks as described in algorithm C
 * This function further updates B' and B'' to remove the partitioned blocks
 * @param bPrime The super block B'
 * @param bDoublePrime  The super block B''
 * @param J The magical constant calculated in Algorithm C determining the length of the new super blocks
 * @return 4-tuple of form (superblock from B', superblock from B'', updated B', updated B'')
 */
std::tuple<std::vector<Block>, std::vector<Block>, std::vector<Block>, std::vector<Block>>
splitSuperBlock(const std::vector<Block> &bPrime, const std::vector<Block> &bDoublePrime, int J) {
    std::vector<Block> newBPrime{};
    std::vector<Block> newBDoublePrime{};
    std::vector<Block> bPrimeSuperBlock{};
    std::vector<Block> bDoublePrimeSuperBlock{};
    int primeCounter = 0;
    int doublePrimeCounter = 0;
    for (const auto &block: bPrime) {
        if (primeCounter < J) {
            bPrimeSuperBlock.push_back(block);
        } else {
            newBPrime.push_back(block);
        }
        if (block.type == "y") {
            primeCounter++;
        }
    }
    for (const auto &block: bDoublePrime) {
        if (doublePrimeCounter < J) {
            bDoublePrimeSuperBlock.push_back(block);
        } else {
            newBDoublePrime.push_back(block);
        }
        if (block.type == "x") {
            doublePrimeCounter++;
        }
    }

    return std::tuple<std::vector<Block>, std::vector<Block>, std::vector<Block>, std::vector<Block>>{bPrimeSuperBlock,
                                                                                                      bDoublePrimeSuperBlock,
                                                                                                      newBPrime,
                                                                                                      newBDoublePrime};
}


std::string generateFoldsFromAlgorithmC(const std::string &sequence) {
    // Decompose the sequence into blocks
    auto blocks = decomposeSequence(sequence);

    // Assign block types correctly
    assignBlockTypes(blocks);

    // Partition the sequence into B' and B''
    std::vector<Block> B_prime, B_double_prime;
    partitionSequence(blocks, B_prime, B_double_prime);

    // Calculate N_x(B'') and N_y(B')
    int N_x_B_double_prime = 0;
    int N_y_B_prime = 0;

    // Calculate N_x(B'')
    for (const auto &block: B_double_prime) {
        if (block.type == "x") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_x_B_double_prime++;
                }
            }
        }
    }

    // Calculate N_y(B')
    for (const auto &block: B_prime) {
        if (block.type == "y") {
            for (char residue: block.residues) {
                if (residue == 'h') {
                    N_y_B_prime++;
                }
            }
        }
    }

    // Step 2: Determine K and J
    int X = N_x_B_double_prime;
    int Y = N_y_B_prime;
    int min_Nx_Ny = std::min(X, Y);

    // Define f(x)=sqrt(x)
    int K = static_cast<int>(std::floor(std::sqrt(min_Nx_Ny)));
    int L = min_Nx_Ny;

    // Calculate J
    int J = static_cast<int>(std::floor(static_cast<double>(L - 2 * K + 1) / K));

    // If K == 1 or J < 2, apply Algorithm B
    if (K == 1 || J < 2) {
        std::cout << "Applying Algorithm B instead." << std::endl;
        return applyAlgorithmB(B_prime, B_double_prime, sequence);
    }

    // TODO: Be clear in exactly what "super block of length J means" - it cannot possibly mean sequence length (too small)
    // Something along the lines of K * J = full_sequence
    // For now we will assume that J is the number of X or Y blocks in a superblock (whichever is... largest?)
    std::vector<Block> remainingBPrime;
    remainingBPrime.assign(B_prime.begin(), B_prime.end());
    std::vector<Block> remainingBDoublePrime;
    remainingBDoublePrime.assign(B_double_prime.begin(), B_double_prime.end());
    std::string totalFolds = "";

    // We try a slightly different approach, fold B'/B'' in one go, and connect them, with number of layers = K
    totalFolds += foldSuperBlock(B_prime, J, K, true);
    // A bit unclear flag argument here, but foldUp is only ever false for 2nd part of superblock, and in that case
    // the first block will be a seperator, so we can handle connection only in this part
    totalFolds += foldSuperBlock(B_double_prime, J, K, false);

    return totalFolds;
}

void foldInModel(const std::string &sequence, const std::string &folds, int cutoff) {
    std::string newSequence;
    std::string newFolds;
    if (cutoff == sequence.length()) {
        newSequence = sequence;
        newFolds = folds;
    } else {
        newSequence = sequence.substr(0, cutoff);
        newFolds = folds.substr(0, cutoff - 1);
    }
    std::cout << newFolds << "\n";
    std::cout << "Last fold: " << newFolds.substr(newFolds.length() - 1, 1) << "\n";
    try {
        HPModel model(newSequence, newFolds);
        int score = model.getScore();
        std::cout << "Number of hydrophobic contacts: " << score << std::endl;

        // Optional: Output the points for visualization
        std::string jsonPoints = model.pointsToJson();
        std::cout << "Protein structure: " << jsonPoints << std::endl;

        Application app;
        if (!app.initialize(1000, 800, "HPModel Visualization")) {
            throw std::runtime_error("Failed to initialize visualization.");
        }
        app.initializeFromHPModel(model);
        app.run();
    } catch (const std::exception &e) {
        std::cerr << "Error: " << e.what() << std::endl;
    }
}


int main() {
    // Example sequence to trigger Algorithm C
    // Using two identical blocks, expect symmetry
    std::string expected_superblock_1 = "hppphphpppphppphpphppppph";
    std::string seperator = "pppp";
    std::string expected_superblock_2 = "hppphphpppphppphpphppppph";

    // TESTED SYMMETRIC SEQUENCE FOR 3D FOLD
    std::string sequence =
            expected_superblock_1 + seperator + expected_superblock_2 + seperator + expected_superblock_1 + seperator +
            expected_superblock_2;

    // UNCOMMENT FOR WORKING 2D FOLDING EXAMPLE (S2-1 from paper)
    //sequence = "hphhpphhhhphhhpphhpphphhphphhpphhppphpppppppphh";

    // S2-2 - 2D FOLD Working
    sequence = "hhhhphhphhhhhpphpphhpphpppppphpphppphpphhpphhhph";

    // S2-3 - failing but typo in paper makes the sequence ambiguous??
    //sequence = "phphhphhhhhhpphphpphphhphphppphpphhpphhpphphpphp";

    // S2-4 - 2D fold, works TODO: but unbalanced, not ending where it started
    //sequence = "phphhpphphhhpphhphhppphhhhhpphphphphpppphpphphp";

    // S2-5 - TODO: FAILS, likely hitting a special case that we have not implemented
    //sequence = "pphppphphhhhpphhhhphhphhhpphphphpphpppppphhphhph";

    // S2-6 - 2D fold, TODO: FAILS, seems likely regular fold failure
    //sequence = "hhhppphhphphhphhphhphppppppphphpphppphpphhhhhhph";

    // S2-7 - TODO: FAILS, special case again
    //sequence = "phpppphphhhphphhhhphhphhppphphppphhhpphhpphhppph";

    // S2-8 - TODO: FAILS, special case again
    //sequence = "phhphhhphhhhpphhhpppppphphhpphhphppphhphphphhppp";

    // S2-9 - 2D fold TODO: Unabalanced
    //sequence = "phphpppphphphpphphhhhhhpphhhphpphphhpphphhhpppph";

    // S2-10 - 2D fold TODO: FAILS, severely broken folding
    //sequence = "phhpppppphhppphhhphpphphhpphpphpphhpphhhhhhhpphh";


    std::cout << "SEQUENCE LENGTH " << sequence.length() << "\n";

    // Generate folds using Algorithm C
    std::string folds = generateFoldsFromAlgorithmC(sequence);
    // TODO: DEBUG REMOVE
    for (int i = folds.length(); i < sequence.length() - 1; ++i) {
        folds += "f";
    }

    // Validate folds length
    if (folds.length() != sequence.length() - 1) {
        std::cerr << "Error: Number of folds (" << folds.length()
                  << ") does not match expected (" << sequence.length() - 1 << ")."
                  << std::endl;
        return -1;
    }

    foldInModel(sequence, folds, sequence.length());
    //foldInModel(sequence, folds, 15);

    return 0;
}