#include <glad/glad.h>
#include "Application.h"
#include <GLFW/glfw3.h>
#include <glm.hpp>
#include <gtc/matrix_transform.hpp>
#include <gtc/type_ptr.hpp>
#include "imgui.h"
#include "imgui_impl_glfw.h"
#include "imgui_impl_opengl3.h"
#include "InputHandler.h"
#include "Utilities.h"
#include <iostream>
#include <fstream>
#include <sstream>

Application::Application()
    : m_window(nullptr),
      m_width(1000),
      m_height(800),
      m_title("3D Visualization"),
      m_showUI(true),
      m_showGridPoints(false),
      m_firstMouse(true), // Initialize m_firstMouse here
      m_deltaTime(0.0f),
      m_lastFrame(0.0f),
      m_cameraPos(0.0f, 0.0f, 20.0f),
      m_cameraFront(0.0f, 0.0f, -1.0f),
      m_cameraUp(0.0f, 1.0f, 0.0f),
      m_yaw(-90.0f),
      m_pitch(0.0f),
      m_fov(45.0f),
      m_darkMode(true)
{
    memset(m_jsonInput, 0, sizeof(m_jsonInput));
    m_jsonErrorMessage = "";
}

Application::~Application()
{
    cleanupImGui();
    if (m_window)
    {
        glfwDestroyWindow(m_window);
        glfwTerminate();
    }
}

bool Application::initialize(int width, int height, const std::string& title)
{
    m_width = width;
    m_height = height;
    m_title = title;

    if (!glfwInit())
    {
        std::cerr << "Failed to initialize GLFW\n";
        return false;
    }

    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

#ifdef __APPLE__
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif

    m_window = glfwCreateWindow(m_width, m_height, m_title.c_str(), nullptr, nullptr);
    if (!m_window)
    {
        std::cerr << "Failed to create GLFW window\n";
        glfwTerminate();
        return false;
    }

    glfwMakeContextCurrent(m_window);
    glfwSetFramebufferSizeCallback(m_window, framebuffer_size_callback);
    glfwSetCursorPosCallback(m_window, mouse_callback);
    glfwSetScrollCallback(m_window, scroll_callback);

    // Set the user pointer to this instance
    glfwSetWindowUserPointer(m_window, this);

    // Register the key callback with a lambda that captures this instance
    glfwSetKeyCallback(m_window, [](GLFWwindow* window, int key, int scancode, int action, int mods) {
        Application* app = reinterpret_cast<Application*>(glfwGetWindowUserPointer(window));
        if (app)
        {
            // Pass member variables directly to the InputHandler key callback
            InputHandler::key_callback(window, key, scancode, action, mods, app->m_showUI, app->m_firstMouse);
        }
    });

    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cerr << "Failed to initialize GLAD\n";
        return false;
    }

    glEnable(GL_DEPTH_TEST);

    m_renderer = std::make_unique<Renderer>();
    m_renderer->initialize();

    initImGui();
    return true;
}

void Application::initImGui()
{
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    if (m_darkMode) {
        ImGui::StyleColorsDark();
    }
    else {
        ImGui::StyleColorsLight();
    }
    ImGuiIO& io = ImGui::GetIO();
    (void)io;
    ImGuiStyle& style = ImGui::GetStyle();
    style.WindowRounding = 0.0f;
    if (m_darkMode)
        style.Colors[ImGuiCol_WindowBg] = ImVec4(0.1f, 0.1f, 0.1f, 1.0f);
    else
        style.Colors[ImGuiCol_WindowBg] = ImVec4(1.0f, 1.0f, 1.0f, 1.0f);

    ImGui_ImplGlfw_InitForOpenGL(m_window, true);
    ImGui_ImplOpenGL3_Init("#version 330");
}

void Application::setImGuiStyle() {
    if (m_darkMode) {
        ImGui::StyleColorsDark();
    }
    else {
        ImGui::StyleColorsLight();
    }
}

void Application::cleanupImGui()
{
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();
}

void Application::run()
{
    while (!glfwWindowShouldClose(m_window))
    {
        float currentFrame = glfwGetTime();
        m_deltaTime = currentFrame - m_lastFrame;
        m_lastFrame = currentFrame;

        processInput(); // Process real-time input here

        ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
        ImGui::NewFrame();

        if (m_showUI)
        {
            // Render ImGui UI
            ImGui::SetNextWindowPos(ImVec2(0, 0));
            ImGui::SetNextWindowSize(ImVec2(250.0f, (float)m_height));
            ImGui::Begin("Control Panel", nullptr, ImGuiWindowFlags_NoResize | ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse);

            static int spherePos[3] = {0, 0, 0};
            ImGui::InputInt("X Position", &spherePos[0]);
            ImGui::InputInt("Y Position", &spherePos[1]);
            ImGui::InputInt("Z Position", &spherePos[2]);

            static bool highlight = false;
            static bool hhContact = false;
            ImGui::Checkbox("Hydrophobic", &highlight);
            ImGui::Checkbox("Part of H-H Contact", &hhContact);
            if (ImGui::Button("Add Sphere"))
            {
                try {
                    addSphere(spherePos[0], spherePos[1], -spherePos[2], highlight, hhContact);
                }
                catch (const std::exception& e) {
                    m_errorMessage = e.what();
                }
            }
            if (!m_errorMessage.empty())
            {
                ImGui::TextColored(ImVec4(1.0f, 0.0f, 0.0f, 1.0f), "%s", m_errorMessage.c_str());
            }

            if (ImGui::Button("Toggle Theme"))
            {
                m_darkMode = !m_darkMode;
                setImGuiStyle();
            }
           ImGui::Separator();
            if (ImGui::CollapsingHeader("Spheres List", ImGuiTreeNodeFlags_DefaultOpen)) {
                // Only show the table if the header is open
                if (ImGui::BeginTable("SpheresTable", 3, ImGuiTableFlags_Borders | ImGuiTableFlags_RowBg)) {
                    ImGui::TableSetupColumn("Coordinates");
                    ImGui::TableSetupColumn("Visible?");
                    ImGui::TableSetupColumn("Delete?");
                    ImGui::TableHeadersRow();

                    std::vector<std::tuple<int, int, int> > toDelete;

                    // Instead of iterating over map in arbitrary order,
                    // we will iterate over the sphereSequence (which is insertion order).
                    for (auto &coords: m_renderer->getSphereSequence()) {
                        auto it = m_renderer->getSpheres().find(coords);
                        if (it == m_renderer->getSpheres().end())
                            continue;

                        Sphere &sphere = it->second;
                        int x = std::get<0>(coords);
                        int y = std::get<1>(coords);
                        int z = std::get<2>(coords);

                        ImGui::TableNextRow();

                        // Coordinates column
                        ImGui::TableSetColumnIndex(0);
                        ImGui::Text("(%d, %d, %d)", x, y, z);

                        // Visible? (checkbox)
                        ImGui::TableSetColumnIndex(1);
                        bool isVisible = !sphere.hidden;
                        std::string checkboxID = "##Visible_" + std::to_string(x) + "_" +
                                                 std::to_string(y) + "_" + std::to_string(z);
                        if (ImGui::Checkbox(checkboxID.c_str(), &isVisible)) {
                            sphere.hidden = !isVisible;
                            m_renderer->updateLines(m_renderer->getSphereSequence());
                        }

                        // Delete?
                        ImGui::TableSetColumnIndex(2);
                        std::string deleteButtonID = "Delete##" + std::to_string(x) + "_" +
                                                     std::to_string(y) + "_" + std::to_string(z);
                        if (ImGui::Button(deleteButtonID.c_str())) {
                            toDelete.push_back(coords);
                        }
                    }

                    // Perform deletions
                    for (auto &coords: toDelete) {
                        // Remove from map
                        m_renderer->getSpheres().erase(coords);

                        // Remove from sphereSequence
                        auto &seq = m_renderer->getSphereSequence();
                        seq.erase(std::remove(seq.begin(), seq.end(), coords), seq.end());

                        // Update lines
                        m_renderer->updateLines(seq);
                    }

                    ImGui::EndTable();
                }
            }

            ImGui::Checkbox("Show Grid Points", &m_showGridPoints);
            m_renderer->toggleGridPoints(m_showGridPoints);

            ImGui::Text("JSON Input:");
            ImGui::InputTextMultiline("##jsonInput", m_jsonInput, sizeof(m_jsonInput), ImVec2(-1.0f, 150.0f));

            if (ImGui::Button("Submit JSON"))
            {
                initializeRendererWithModel(m_jsonInput);
            }

            if (!m_jsonErrorMessage.empty())
            {
                ImGui::TextColored(ImVec4(1.0f, 0.0f, 0.0f, 1.0f), "%s", m_jsonErrorMessage.c_str());
            }

            ImGui::End();
        }

        update(); // Update the application state

        // Render 3D scene
        if (m_darkMode) {
            // Dark background
            glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
        }
        else {
            // Light background
            glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
        }
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glm::mat4 view = glm::lookAt(m_cameraPos, m_cameraPos + m_cameraFront, m_cameraUp);
        glm::mat4 projection = glm::perspective(glm::radians(m_fov), (float)m_width / (float)m_height, 0.1f, 100.0f);
        m_renderer->render(view, projection);

        // Render ImGui
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        glfwSwapBuffers(m_window);
        glfwPollEvents();
    }
}

void Application::processInput()
{
    InputHandler::processInput(m_window, m_deltaTime, m_cameraPos, m_cameraFront, m_cameraUp);
}

void Application::update()
{
}

void Application::loadModelFromJson(const std::string& jsonPath)
{
    std::ifstream file(jsonPath);
    if (file.is_open())
    {
        std::stringstream buffer;
        buffer << file.rdbuf();
        std::string jsonContent = buffer.str();
        initializeRendererWithModel(jsonContent);
    }
    else
    {
        std::cerr << "Failed to open JSON file: " << jsonPath << "\n";
    }
}

void Application::addSphere(int x, int y, int z, bool hydrophobic, bool hhContact)
{
    Sphere sphere;
    sphere.position = glm::vec3(x, y, z);
    if (hydrophobic && hhContact)
    {
        throw std::runtime_error("A sphere cannot be both hydrophobic and part of an H-H contact.");
    }
        if (hhContact)
        sphere.color = glm::vec3(0.0f, 0.0f, 1.0f);
    else if (hydrophobic)
        sphere.color = glm::vec3(1.0f, 0.0f, 0.0f);
    else
        sphere.color = glm::vec3(0.5f, 0.5f, 0.5f);

    auto sphereKey = std::make_tuple(x, y, z);
    m_renderer->getSpheres()[sphereKey] = sphere;
    m_renderer->getSphereSequence().push_back(sphereKey);
    m_renderer->updateLines(m_renderer->getSphereSequence());
}

void Application::initializeRendererWithModel(const std::string& jsonStr)
{
    m_renderer->getSpheres().clear();
    m_renderer->getSphereSequence().clear();
    if (!parseHPModelJson(jsonStr, m_renderer->getSpheres(), m_renderer->getSphereSequence(), m_jsonErrorMessage))
    {
        std::cerr << "Error parsing JSON: " << m_jsonErrorMessage << "\n";
    }
    else
    {
        m_renderer->updateLines(m_renderer->getSphereSequence());
    }
}

void Application::framebuffer_size_callback(GLFWwindow* window, int width, int height)
{
    glViewport(0, 0, width, height);
}

void Application::mouse_callback(GLFWwindow* window, double xpos, double ypos)
{
    if (glfwGetInputMode(window, GLFW_CURSOR) != GLFW_CURSOR_DISABLED)
        return;
    static float lastX = 0.0f;
    static float lastY = 0.0f;
    static bool firstMouse = true;
    if (firstMouse)
    {
        lastX = (float)xpos;
        lastY = (float)ypos;
        firstMouse = false;
    }
    float xoffset = (float)xpos - lastX;
    float yoffset = lastY - (float)ypos;
    lastX = (float)xpos;
    lastY = (float)ypos;
    float sensitivity = 0.1f;
    xoffset *= sensitivity;
    yoffset *= sensitivity;
    Application* app = reinterpret_cast<Application*>(glfwGetWindowUserPointer(window));
    if (!app)
        return;
    app->m_yaw += xoffset;
    app->m_pitch += yoffset;
    if (app->m_pitch > 89.0f)
        app->m_pitch = 89.0f;
    if (app->m_pitch < -89.0f)
        app->m_pitch = -89.0f;
    glm::vec3 front;
    front.x = cos(glm::radians(app->m_yaw)) * cos(glm::radians(app->m_pitch));
    front.y = sin(glm::radians(app->m_pitch));
    front.z = sin(glm::radians(app->m_yaw)) * cos(glm::radians(app->m_pitch));
    app->m_cameraFront = glm::normalize(front);
}

void Application::scroll_callback(GLFWwindow* window, double xoffset, double yoffset)
{
    Application* app = reinterpret_cast<Application*>(glfwGetWindowUserPointer(window));
    if (!app)
        return;
    app->m_fov -= (float)yoffset;
    if (app->m_fov < 1.0f)
        app->m_fov = 1.0f;
    if (app->m_fov > 90.0f)
        app->m_fov = 90.0f;
}

void Application::initializeFromHPModel(const HPModel& model) {
    try {
        // Get JSON representation of the model points
        std::string jsonPoints = model.pointsToJson();

        // Clear current visualization
        m_renderer->getSpheres().clear();
        m_renderer->getSphereSequence().clear();

        // Parse JSON into spheres and sequence
        std::string errorMessage;
        if (!parseHPModelJson(jsonPoints, m_renderer->getSpheres(), m_renderer->getSphereSequence(), errorMessage)) {
            throw std::runtime_error("Failed to parse JSON: " + errorMessage);
        }

        // Update visualization
        m_renderer->updateLines(m_renderer->getSphereSequence());
    } catch (const std::exception& e) {
        std::cerr << "Error initializing visualization: " << e.what() << std::endl;
    }
}