#include "Renderer.h"
#include <gtc/matrix_transform.hpp>
#include <iostream>
#include <Utilities.h>
#include <glad/glad.h>

// Constructor
Renderer::Renderer()
    : shader(nullptr), sphereVAO(0), sphereVBO(0), sphereEBO(0),
      lineVAO(0), lineVBO(0), axesVAO(0), axesVBO(0),
      gridVAO(0), gridVBO(0), showGridPointsFlag(false),
      sphereScale(1.0f), lineThickness(3.0f)
{}

// Destructor
Renderer::~Renderer()
{
    // Clean up resources
    if(shader) delete shader;
    glDeleteVertexArrays(1, &sphereVAO);
    glDeleteBuffers(1, &sphereVBO);
    glDeleteBuffers(1, &sphereEBO);
    glDeleteVertexArrays(1, &lineVAO);
    glDeleteBuffers(1, &lineVBO);
    glDeleteVertexArrays(1, &axesVAO);
    glDeleteBuffers(1, &axesVBO);
    glDeleteVertexArrays(1, &gridVAO);
    glDeleteBuffers(1, &gridVBO);
}

// Initialize Renderer
void Renderer::initialize()
{
    // Initialize shader
    shader = new Shader("shaders/vertex_shader.glsl", "shaders/fragment_shader.glsl");

    // Generate sphere data
    SphereGenerator::generateSphere(sphereVertices, sphereIndices, 0.1f, 36, 18);
    setupSphere();

    // Setup lines
    setupLines();

    // Setup axes
    setupAxes();

    // Setup grid
    setupGrid();
}

// Setup Sphere VAO, VBO, EBO
void Renderer::setupSphere()
{
    glGenVertexArrays(1, &sphereVAO);
    glGenBuffers(1, &sphereVBO);
    glGenBuffers(1, &sphereEBO);

    glBindVertexArray(sphereVAO);

    glBindBuffer(GL_ARRAY_BUFFER, sphereVBO);
    glBufferData(GL_ARRAY_BUFFER, sphereVertices.size() * sizeof(float), &sphereVertices[0], GL_STATIC_DRAW);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphereEBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sphereIndices.size() * sizeof(unsigned int), &sphereIndices[0], GL_STATIC_DRAW);

    // Position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Normal attribute
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(1);

    glBindVertexArray(0);
}

// Setup Lines VAO and VBO
void Renderer::setupLines()
{
    glGenVertexArrays(1, &lineVAO);
    glGenBuffers(1, &lineVBO);

    glBindVertexArray(lineVAO);

    glBindBuffer(GL_ARRAY_BUFFER, lineVBO);
    glBufferData(GL_ARRAY_BUFFER, 0, nullptr, GL_DYNAMIC_DRAW);

    // Position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Color attribute
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);

    glBindVertexArray(0);
}

// Setup Axes VAO and VBO
void Renderer::setupAxes()
{
    const float AXIS_LENGTH = 10.0f;

    std::vector<float> axesVertices = {
        // X-axis (Red)
        -AXIS_LENGTH, 0.0f, 0.0f,  1.0f, 0.0f, 0.0f,
         AXIS_LENGTH, 0.0f, 0.0f,  1.0f, 0.0f, 0.0f,
        // Y-axis (Green)
        0.0f, -AXIS_LENGTH, 0.0f,  0.0f, 1.0f, 0.0f,
        0.0f,  AXIS_LENGTH, 0.0f,  0.0f, 1.0f, 0.0f,
        // Z-axis (Blue)
        0.0f, 0.0f,  AXIS_LENGTH,  0.0f, 0.0f, 1.0f,
        0.0f, 0.0f, -AXIS_LENGTH,  0.0f, 0.0f, 1.0f,
    };

    glGenVertexArrays(1, &axesVAO);
    glGenBuffers(1, &axesVBO);

    glBindVertexArray(axesVAO);
    glBindBuffer(GL_ARRAY_BUFFER, axesVBO);
    glBufferData(GL_ARRAY_BUFFER, axesVertices.size() * sizeof(float), axesVertices.data(), GL_STATIC_DRAW);

    // Position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Color attribute
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);

    glBindVertexArray(0);
}

// Setup Grid VAO and VBO
void Renderer::setupGrid()
{
    // Generate grid points at integer coordinates within a range
    std::vector<float> gridPts;
    int gridRange = 10; // Adjust as needed

    for (int x = -gridRange; x <= gridRange; ++x)
    {
        for (int y = -gridRange; y <= gridRange; ++y)
        {
            for (int z = -gridRange; z <= gridRange; ++z)
            {
                gridPts.push_back(static_cast<float>(x));
                gridPts.push_back(static_cast<float>(y));
                gridPts.push_back(static_cast<float>(z));

                // Color (white)
                gridPts.push_back(1.0f); // R
                gridPts.push_back(1.0f); // G
                gridPts.push_back(1.0f); // B
            }
        }
    }

    gridPoints = gridPts;

    glGenVertexArrays(1, &gridVAO);
    glGenBuffers(1, &gridVBO);

    glBindVertexArray(gridVAO);
    glBindBuffer(GL_ARRAY_BUFFER, gridVBO);
    glBufferData(GL_ARRAY_BUFFER, gridPoints.size() * sizeof(float), gridPoints.data(), GL_STATIC_DRAW);

    // Position attribute
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(0);

    // Color attribute
    glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 6 * sizeof(float), (void*)(3 * sizeof(float)));
    glEnableVertexAttribArray(2);

    glBindVertexArray(0);
}

// Update Lines based on sphere sequence
void Renderer::updateLines(const std::vector<std::tuple<int, int, int>>& sphereSeq)
{
    lineVertices.clear();

    if (sphereSeq.size() < 2)
        return;

    for (size_t i = 0; i < sphereSeq.size() - 1; ++i)
    {
        auto currentKey = sphereSeq[i];
        auto nextKey = sphereSeq[i + 1];

        auto currentIt = spheres.find(currentKey);
        auto nextIt = spheres.find(nextKey);

        if (currentIt == spheres.end() || nextIt == spheres.end())
            continue;

        const Sphere& currentSphere = currentIt->second;
        const Sphere& nextSphere = nextIt->second;

        if (currentSphere.hidden || nextSphere.hidden)
            continue; // Skip line if either sphere is hidden

        // Get integer positions for adjacency check
        int x1 = std::get<0>(currentKey);
        int y1 = std::get<1>(currentKey);
        int z1 = std::get<2>(currentKey);

        int x2 = std::get<0>(nextKey);
        int y2 = std::get<1>(nextKey);
        int z2 = std::get<2>(nextKey);

        // Check if they're adjacent by 1
        int xDiff = std::abs(x2 - x1);
        int yDiff = std::abs(y2 - y1);
        int zDiff = std::abs(z2 - z1);

        // Only add line if sum of diffs == 1
        if (xDiff + yDiff + zDiff != 1)
            continue;

        lineVertices.insert(lineVertices.end(), {
            currentSphere.position.x, currentSphere.position.y, currentSphere.position.z, 0.5f, 0.5f, 0.5f,
            nextSphere.position.x,    nextSphere.position.y,    nextSphere.position.z,    0.5f, 0.5f, 0.5f
        });
    }

    glBindBuffer(GL_ARRAY_BUFFER, lineVBO);
    glBufferData(GL_ARRAY_BUFFER, lineVertices.size() * sizeof(float), lineVertices.data(), GL_DYNAMIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}


// Set Sphere Scale
void Renderer::setSphereScale(float scale)
{
    sphereScale = scale;
}

// Set Line Thickness
void Renderer::setLineThickness(float thickness)
{
    lineThickness = thickness;
}

// Toggle Grid Points Visibility
void Renderer::toggleGridPoints(bool show)
{
    showGridPointsFlag = show;
}

// Render Function
void Renderer::render(const glm::mat4& view, const glm::mat4& projection)
{
    shader->use();
    shader->setMat4("view", view);
    shader->setMat4("projection", projection);

    // Render axes
    glBindVertexArray(axesVAO);
    glm::mat4 model = glm::mat4(1.0f);
    shader->setMat4("model", model);
    shader->setBool("useUniformColor", false);
    glDrawArrays(GL_LINES, 0, 6);
    glBindVertexArray(0);

    // Render grid points if enabled
    if(showGridPointsFlag)
    {
        glBindVertexArray(gridVAO);
        glPointSize(5.0f);
        shader->setMat4("model", model);
        shader->setBool("useUniformColor", false);
        glDrawArrays(GL_POINTS, 0, static_cast<unsigned int>(gridPoints.size() / 6));
        glBindVertexArray(0);
    }

    // Render spheres
    glBindVertexArray(sphereVAO);
    shader->setBool("useUniformColor", true);
    for (const auto& [key, sphere] : spheres)
    {
        if (sphere.hidden)
            continue; // Skip hidden spheres

        glm::mat4 model = glm::mat4(1.0f);
        model = glm::translate(model, sphere.position);
        model = glm::scale(model, glm::vec3(sphereScale));

        shader->setMat4("model", model);
        shader->setVec3("sphereColor", sphere.color);

        glDrawElements(GL_TRIANGLES, static_cast<unsigned int>(sphereIndices.size()), GL_UNSIGNED_INT, 0);
    }
    glBindVertexArray(0);

    // -- Render Lines (only if both spheres in a pair are NOT hidden) --
    if (!lineVertices.empty())
    {
        glLineWidth(lineThickness);
        glBindVertexArray(lineVAO);
        shader->setMat4("model", glm::mat4(1.0f));
        shader->setBool("useUniformColor", false);
        glDrawArrays(GL_LINES, 0, static_cast<unsigned int>(lineVertices.size() / 6));
        glBindVertexArray(0);
        glLineWidth(1.0f);
    }
}

