#include "Utilities.h"
#include <glm.hpp>
#include <gtc/matrix_transform.hpp>
#include <iostream>
#include <cctype>
#include "Renderer.h"
#include <HPModel.h>
// Implement WorldToScreen
bool WorldToScreen(const glm::vec3& worldPos, const glm::mat4& view,
                  const glm::mat4& projection,
                  float screenWidth, float screenHeight,
                  float& screenX, float& screenY)
{
    glm::vec4 clipSpacePos = projection * view * glm::vec4(worldPos, 1.0f);

    if (clipSpacePos.w <= 0.0f)
        return false; // Behind the camera

    glm::vec3 ndc = glm::vec3(clipSpacePos) / clipSpacePos.w;

    if (ndc.x < -1.0f || ndc.x > 1.0f || ndc.y < -1.0f || ndc.y > 1.0f)
        return false; // Outside of screen

    // Map to window coordinates
    screenX = (ndc.x + 1.0f) * screenWidth / 2.0f;
    screenY = (1.0f - ndc.y) * screenHeight / 2.0f; // Invert Y for window coordinates

    return true;
}

// Implement a simple JSON parser or use a library like nlohmann/json
// For simplicity, using a basic parser as in your original code

bool parseHPModelJson(const std::string& jsonInput,
                      std::map<std::tuple<int, int, int>, Sphere>& spheres,
                      std::vector<std::tuple<int, int, int>>& sphereSequence,
                      std::string& errorMessage)
{
    // Remove all whitespace
    std::string input;
    for (char c : jsonInput)
    {
        if (!isspace(static_cast<unsigned char>(c)))
            input += c;
    }

    if (input.empty())
    {
        errorMessage = "Input is empty.";
        return false;
    }

    if (input.front() != '[' || input.back() != ']')
    {
        errorMessage = "JSON must start with '[' and end with ']'!";
        return false;
    }

    input = input.substr(1, input.size() - 2); // Remove '[' and ']'

    size_t pos = 0;
    while (pos < input.size())
    {
        if (input[pos] != '{')
        {
            errorMessage = "Expected '{' at position " + std::to_string(pos);
            return false;
        }

        size_t endPos = input.find('}', pos);
        if (endPos == std::string::npos)
        {
            errorMessage = "Expected '}' after position " + std::to_string(pos);
            return false;
        }

        std::string objStr = input.substr(pos + 1, endPos - pos - 1); // Get content inside '{' and '}'

        // Parse the object string
        int x = 0, y = 0, z = 0;
        bool hydrophobic = false, partOfScoringPair = false;
        bool xSet = false, ySet = false, zSet = false, hydroSet = false, scoringPairSet = false;

        size_t objPos = 0;
        while (objPos < objStr.size())
        {
            if (objStr[objPos] != '"')
            {
                errorMessage = "Expected '\"' for key at position " + std::to_string(objPos);
                return false;
            }
            size_t keyEnd = objStr.find('"', objPos + 1);
            if (keyEnd == std::string::npos)
            {
                errorMessage = "Expected '\"' to end key at position " + std::to_string(objPos + 1);
                return false;
            }

            std::string key = objStr.substr(objPos + 1, keyEnd - objPos - 1);

            size_t colonPos = objStr.find(':', keyEnd + 1);
            if (colonPos == std::string::npos)
            {
                errorMessage = "Expected ':' after key \"" + key + "\"";
                return false;
            }

            size_t valueStart = colonPos + 1;

            if (key == "hydrophobic")
            {
                if (objStr.compare(valueStart, 4, "true") == 0)
                {
                    hydrophobic = true;
                }
                else if (objStr.compare(valueStart, 5, "false") == 0)
                {
                    hydrophobic = false;
                }
                else
                {
                    errorMessage = "Invalid value for key \"hydrophobic\"";
                    return false;
                }
                hydroSet = true;
            }
            else if (key == "partOfScoringPair")
            {
                if (objStr.compare(valueStart, 4, "true") == 0)
                {
                    partOfScoringPair = true;
                }
                else if (objStr.compare(valueStart, 5, "false") == 0)
                {
                    partOfScoringPair = false;
                }
                else
                {
                    errorMessage = "Invalid value for key \"partOfScoringPair\"";
                    return false;
                }
                scoringPairSet = true;
            }
            else
            {
                // Parse integer values for x, y, z
                int value = 0;
                size_t valueEnd = valueStart;
                while (valueEnd < objStr.size() &&
                       (isdigit(objStr[valueEnd]) || objStr[valueEnd] == '-'))
                {
                    valueEnd++;
                }
                std::string valueStr = objStr.substr(valueStart, valueEnd - valueStart);
                try
                {
                    value = std::stoi(valueStr);
                }
                catch (std::exception&)
                {
                    errorMessage = "Invalid integer value for key \"" + key + "\"";
                    return false;
                }

                if (key == "x")
                {
                    x = value;
                    xSet = true;
                }
                else if (key == "y")
                {
                    y = value;
                    ySet = true;
                }
                else if (key == "z")
                {
                    z = value;
                    zSet = true;
                }
                else
                {
                    errorMessage = "Unexpected key \"" + key + "\"";
                    return false;
                }
            }

            objPos = objStr.find(',', colonPos + 1);
            if (objPos == std::string::npos)
                break;
            objPos++;
        }

        if (!xSet || !ySet || !zSet)
        {
            errorMessage = "Missing key(s) in object!";
            return false;
        }

        // Add sphere to the map
        Sphere sphere;
        sphere.position = glm::vec3(x, y, -z); // Adjust Z-axis as in your code
        if (scoringPairSet && partOfScoringPair) {
            sphere.color = glm::vec3(0.0f, 0.0f, 1.0f); // Blue for scoring pairs
        } else if (hydrophobic) {
            sphere.color = glm::vec3(1.0f, 0.0f, 0.0f); // Red for hydrophobic
        } else {
            sphere.color = glm::vec3(0.5f, 0.5f, 0.5f); // Gray for polar
        }

        auto sphereKey = std::make_tuple(x, y, z);
        spheres[sphereKey] = sphere;

        // Add to sphereSequence for sequential linking
        sphereSequence.push_back(sphereKey);

        pos = endPos + 1;
        if (pos < input.size() && input[pos] == ',')
        {
            pos++;
        }
    }

    return true;
}

// Function to initialize visualization from JSON points using your parser
void visualizeHPModel(Renderer* renderer, const std::string& jsonPoints) {
    // Map and sequence to store parsed spheres and their connections
    std::map<std::tuple<int, int, int>, Sphere> spheres;
    std::vector<std::tuple<int, int, int>> sphereSequence;
    std::string errorMessage;

    // Parse the JSON string
    if (!parseHPModelJson(jsonPoints, spheres, sphereSequence, errorMessage)) {
        std::cerr << "Error parsing JSON for visualization: " << errorMessage << std::endl;
        return;
    }

    // Clear current renderer data
    renderer->getSpheres().clear();
    renderer->getSphereSequence().clear();

    // Add parsed data to the renderer
    renderer->getSpheres() = spheres;
    renderer->getSphereSequence() = sphereSequence;

    // Update the renderer lines based on the new sphere sequence
    renderer->updateLines(renderer->getSphereSequence());

    std::cout << "Visualization initialized successfully from JSON." << std::endl;
}

void visualizeHPModel(Renderer* renderer, const HPModel& model) {
    try {
        // Get the JSON representation of the folded protein structure
        std::string jsonPoints = model.pointsToJson();

        // Clear existing visualization data
        renderer->getSpheres().clear();
        renderer->getSphereSequence().clear();

        // Parse JSON input and update the renderer's data
        std::string errorMessage;
        if (!parseHPModelJson(jsonPoints, renderer->getSpheres(), renderer->getSphereSequence(), errorMessage)) {
            throw std::runtime_error("Failed to parse JSON input: " + errorMessage);
        }

        // Update lines and other visualization elements
        renderer->updateLines(renderer->getSphereSequence());
    } catch (const std::exception& e) {
        std::cerr << "Error in visualizeHPModel: " << e.what() << std::endl;
    }
}