#include <iostream>
#include <utility>
#include <vector>
#include "../cache/FileCacheManager.h"
#include "../../model/HPModel.h"
#include "BruteForce.h"
#include <stdexcept>
#include <algorithm>

std::string validateSequence(const std::string &sequence) {
    std::string sanitized_sequence;
    for (char c: sequence) {
        char lower_c;
        if (c == 'H') {
            lower_c = 'h';
        } else if (c == 'P') {
            lower_c = 'p';
        } else {
            lower_c = c;
        }
        if (lower_c != 'h' && lower_c != 'p') {
            throw std::invalid_argument("Invalid character found in sequence, '" + std::to_string(c) +
                                        "' is not a valid char, only 'h' and 'p' allowed");
        }
        sanitized_sequence += lower_c;
    }
    return sanitized_sequence;
}

void generateAllFolds(std::vector<std::string> &acc, int length, std::string prefix) {
    std::string motifs = "frlud";

    if (length == 0) {
        acc.push_back(prefix);
        return;
    }

    for (int i = 0; i < motifs.size(); i++) {
        std::string newPrefix;

        newPrefix = prefix + motifs[i];

        generateAllFolds(acc, length - 1, newPrefix);
    }
}

std::vector<std::string> generateAllFolds(const std::string &sequence) {
    std::string sanitizedSequence = validateSequence(sequence);
    auto numberOfFolds = sanitizedSequence.size() - 1;
    std::vector<std::string> allPossibleFolds;
    generateAllFolds(allPossibleFolds, numberOfFolds, "");
    return allPossibleFolds;
}

std::tuple<int, std::string> findOptimalScore(const std::string &sequence) {
    std::string sanitized_sequence = validateSequence(sequence);

    auto numberOfFolds = sequence.size() - 1;
    std::vector<std::string> allPossibleFolds;
    generateAllFolds(allPossibleFolds, numberOfFolds, "");
    int bestScore = 0;
    std::string bestFold;
    for (auto fold: allPossibleFolds) {
        try {
            auto model = HPModel(sanitized_sequence, fold);
            int score = model.getScore();
            if (score > bestScore) {
                bestScore = score;
                bestFold = fold;
            }
        } catch (...) {
            continue;
        }
    }
    return {bestScore, bestFold};
}

void executeAndLog(const std::string &sequence) {
    Executor executor = Executor(sequence);
    executor.execute();
    auto data = executor.data;
    saveSequence(sequence, OPTIMAL_FOUND, BRUTE_FORCE, data.bestFold, data.bestScore);
}


void executeToDos() {
    auto todos = getSequencesToDo();
    for (auto todo: todos) {
        executeAndLog(todo);
    }
}

Executor::Executor(const std::string &_sequence) : sequence(_sequence) {
    std::cout << "Creating executor\n";
    auto foundData = loadSequence(sequence, BRUTE_FORCE);
    this->data = foundData;
    // Generating all possible folds is so relatively simple that it does not make sense to store this data separately
    // We can just generate it on startup
    std::cout << "Generating folds\n";
    this->possibleFolds = generateAllFolds(sequence);
    std::cout << "Finished generating folds\n";
    this->expectedFoldLength = sequence.size() - 1;
    this->cache = std::map<std::string, HPModel>{};
    this->impossibleFoldCache = std::map<int, std::vector<std::string>>{};
    for (const auto& impossibleFold : data.impossiblePrefixes) {
        auto keyFind = impossibleFoldCache.find((int) impossibleFold.size());
        if (keyFind != impossibleFoldCache.end()) {
            auto folds = keyFind->second;
            folds.push_back(impossibleFold);
            impossibleFoldCache.insert({(int) impossibleFold.size(), folds});
        } else {
            impossibleFoldCache.insert({(int) impossibleFold.size(), std::vector<std::string>{impossibleFold}});
        }
    }

}


// Note that we assume that the internal cache data is always up-to-date!
void Executor::logProgress() {
    std::cout << "LOGGING PROGRESS\n";
    saveSequence(
            sequence,
            BRUTE_FORCE,
            data.bestScore,
            data
    );
}

void Executor::addImpossibleFold(std::string fold) {
    int size = (int) fold.size();
    auto stringVectorIt = impossibleFoldCache.find(size);
    if (stringVectorIt != impossibleFoldCache.end()) {
        auto impossibleFolds = stringVectorIt->second;
        impossibleFolds.push_back(fold);
        impossibleFoldCache.insert({size, impossibleFolds});
    } else {
        impossibleFoldCache.insert({size, std::vector<std::string>{fold}});
    }
}

DIRECTION convertStringToDirection(std::string motif) {
    if (motif == "f") {
        return FORWARD;
    }
    if (motif == "r") {
        return RIGHT;
    }
    if (motif == "l") {
        return LEFT;
    }
    if (motif == "u") {
        return UP;
    }
    if (motif == "d") {
        return DOWN;
    }

    throw std::invalid_argument("UNKNOWN MOTIF " + motif + "\n");
}

void Executor::updateModel(HPModel newModel, std::string subFold) {
    cache.insert({subFold, newModel});
}

void Executor::execute() {
    int currentIteration = 0;
    if (!data.isEmpty()) {
        currentIteration = data.currentFoldLength;
    }
    for (int i = currentIteration; i < expectedFoldLength; ++i) {
        std::cout << "CURRENT ITERATION: " << i << "\n";
        int bestScoreSoFar = 0;
        std::string bestFoldSoFar = "";
        std::vector<std::string> newImpossibleFolds{};
        std::vector<std::string> newFoldsChecked{};

        for (const auto& fold : possibleFolds) {
            // For i = 0 there cannot possibly be anything in the cache
            // Note that substr(0, i) looks for substrings exactly one character smaller than what we have currently
            // This means that we are looking at the substring excluding the current fold!
            std::string currentFold = fold.substr(0, i + 1);
            std::string motif = fold.substr(i, 1);
            DIRECTION foldDirection = convertStringToDirection(motif);
            newFoldsChecked.push_back(currentFold);

            if (i > 1) {
                std::string foldSubStr = fold.substr(0, i);

                // Check if the fold is known to be impossible
                auto impossibleFind = impossibleFoldCache.find((int) foldSubStr.size());
                if (impossibleFind != impossibleFoldCache.end()) {
                    auto knownImpossibleStrings = impossibleFind->second;
                    if (std::find(knownImpossibleStrings.begin(), knownImpossibleStrings.end(), foldSubStr) != knownImpossibleStrings.end()) {
                        addImpossibleFold(currentFold);
                        continue;
                    }
                } else {
                    impossibleFoldCache.insert({(int) foldSubStr.size(), std::vector<std::string> {}});
                }

                // Look up the model in the cache, if a model exists we can skip doing all the previous folds,
                // otherwise we have to do the hard work again D:
                auto cachedFold = cache.find(foldSubStr);
                if (cachedFold != cache.end()) {
                    HPModel model = cachedFold->second;
                    try {
                        model.fold(foldDirection);
                        updateModel(model, currentFold);
                        int score = model.getScore();
                        // Note that it is important we check >= and not just >, otherwise we cannot guarantee fold length matches sequence length
                        if (score >= bestScoreSoFar) {
                            bestScoreSoFar = score;
                            bestFoldSoFar = currentFold;
                        }
                    } catch (...) {
                        addImpossibleFold(currentFold);
                        continue;
                    }
                } else {
                    try {
                        HPModel model = HPModel(sequence, currentFold);
                        updateModel(model, currentFold);
                        int score = model.getScore();
                        if (score >= bestScoreSoFar) {
                            bestScoreSoFar = score;
                            bestFoldSoFar = currentFold;
                        }
                    } catch (...) {
                        addImpossibleFold(currentFold);
                        continue;
                    }
                }
            }
        }


        // Nothing interesting happens for the first 2 folds, so nothing to log
        if (i > 1) {
            data.currentFoldLength = i;
            if (bestScoreSoFar > data.bestScore) {
                data.bestScore = bestScoreSoFar;
                data.bestFold = bestFoldSoFar;
            }
            data.impossiblePrefixes.insert(data.impossiblePrefixes.end(),
                                           newImpossibleFolds.begin(),
                                           newImpossibleFolds.end());
            data.checkedFolds.insert(data.checkedFolds.end(),
                                     newFoldsChecked.begin(),
                                     newFoldsChecked.end());
            logProgress();
        }
    }
}





